__doc__ = """Snake friction case from X. Zhang et. al. Nat. Comm. 2021"""

""" This environment is constructed based on the PyElastica:ContinuumSnakeCase """

from typing import Optional

from collections import defaultdict
from functools import partial


from gymnasium import spaces
from gymnasium.utils import seeding

from elastica import *

import numpy as np


def compute_projected_velocity(plot_params: dict, period):
    time_per_period = np.array(plot_params["time"]) / period
    avg_velocity = np.array(plot_params["avg_velocity"])
    center_of_mass = np.array(plot_params["center_of_mass"])

    # Compute rod velocity in rod direction. We need to compute that because,
    # after snake starts to move it chooses an arbitrary direction, which does not
    # have to be initial tangent direction of the rod. Thus we need to project the
    # snake velocity with respect to its new tangent and roll direction, after that
    # we will get the correct forward and lateral speed. After this projection
    # lateral velocity of the snake has to be oscillating between + and - values with
    # zero mean.

    # Number of steps in one period.
    period_step = int(1.0 / (time_per_period[-1] - time_per_period[-2]))
    number_of_period = int(time_per_period[-1])
    if number_of_period - 2 <= 0:
        return (0, 0, 0, 0)

    # Center of mass position averaged in one period
    center_of_mass_averaged_over_one_period = np.zeros((number_of_period - 2, 3))
    for i in range(1, number_of_period - 1):
        # position of center of mass averaged over one period
        center_of_mass_averaged_over_one_period[i - 1] = np.mean(
            center_of_mass[(i + 1) * period_step : (i + 2) * period_step]
            - center_of_mass[(i + 0) * period_step : (i + 1) * period_step],
            axis=0,
        )
    # Average the rod directions over multiple periods and get the direction of the rod.
    direction_of_rod = np.mean(center_of_mass_averaged_over_one_period, axis=0)
    direction_of_rod /= np.linalg.norm(direction_of_rod, ord=2)

    # Compute the projected rod velocity in the direction of the rod
    velocity_mag_in_direction_of_rod = np.einsum(
        "ji,i->j", avg_velocity, direction_of_rod
    )
    velocity_in_direction_of_rod = np.einsum(
        "j,i->ji", velocity_mag_in_direction_of_rod, direction_of_rod
    )

    # Get the lateral or roll velocity of the rod after subtracting its projected
    # velocity in the direction of rod
    velocity_in_rod_roll_dir = avg_velocity - velocity_in_direction_of_rod

    # Compute the average velocity over the simulation, this can be used for optimizing snake
    # for fastest forward velocity. We start after first period, because of the ramping up happens
    # in first period.
    average_velocity_over_simulation = np.mean(
        velocity_in_direction_of_rod[period_step * 2 :], axis=0
    )

    return (
        velocity_in_direction_of_rod,
        velocity_in_rod_roll_dir,
        average_velocity_over_simulation[2],
        average_velocity_over_simulation[0],
    )


class SnakeSimulator(BaseSystemCollection, Constraints, Forcing, CallBacks):
    pass


class ContinuumSnakeCallBack(CallBackBaseClass):
    """
    Call back function for continuum snake
    """

    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params["time"].append(time)
            self.callback_params["step"].append(current_step)
            self.callback_params["position"].append(system.position_collection.copy())
            self.callback_params["velocity"].append(system.velocity_collection.copy())
            self.callback_params["avg_velocity"].append(
                system.compute_velocity_center_of_mass()
            )

            self.callback_params["center_of_mass"].append(
                system.compute_position_center_of_mass()
            )
            self.callback_params["curvature"].append(system.kappa.copy())
            return


class ContinuumSnakeEnv(Env):
    metadata = {"render.modes": ["rgb_array", "human"]}

    def __init__(self):
        # Action space
        action_space_low = -np.ones(7) * 1e-2
        action_space_low[-1] = 0.5
        action_space_high = np.ones(7) * 1e-2
        action_space_high[-1] = 3.0
        self.action_space = spaces.Box(
            action_space_low, action_space_high, dtype=np.float32
        )

        # State space
        self._n_elem = 50
        observation_space = (((self._n_elem + 1) * 3 * 2) + ((self._n_elem) * 9),)
        self.observation_space = spaces.Box(
            -np.inf, np.inf, shape=observation_space, dtype=np.float32
        )

        self.metadata = {}

        # Determinism
        self.seed()

        self.viewer = None
        self.rendere = None

    def seed(self, seed=None):
        # Deprecated in new gym
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.snake_sim, self.stepper, self.muscle_torque, self.data = self._build()
        self.time = np.float64(0.0)
        if return_info:
            return self.get_state(), {}
        else:
            return self.get_state()

    def get_state(self):
        # Build state
        rod = self.shearable_rod
        pos_state = rod.position_collection.ravel()
        vel_state = rod.velocity_collection.ravel()
        dir_state = rod.director_collection.ravel()
        return np.concatenate([pos_state, vel_state, dir_state], dtype=np.float32)

    def set_action(self, action):
        wave_length = action[-1]
        b_coeff = action[:-1]
        self.muscle_torque().__init__(
            base_length=self.base_length,
            b_coeff=b_coeff,
            period=self.period,
            wave_number=2.0 * np.pi / (wave_length),
            phase_shift=0.0,
            rest_lengths=self.shearable_rod.rest_lengths,
            ramp_up_time=self.period,
            direction=np.array([0.0, 1.0, 0.0]),
            with_spline=True,
        )

    def step(self, action):
        err_msg = f"{action!r} ({type(action)}) invalid: expected {self.action_space}"
        assert self.action_space.contains(action), err_msg

        """ Set action """
        self.set_action(action)

        """ Run simulation """
        for _ in range(self.step_skip):
            self.time = self.stepper(time=self.time)

        # Compute the average forward velocity. These will be used for optimization.
        [_, _, avg_forward, avg_lateral] = compute_projected_velocity(
            self.data, self.period
        )

        terminated = False
        truncated = False
        if self.time >= self.final_time:
            terminated = True
            truncated = True

        info = {}

        return self.get_state(), avg_forward, terminated, truncated, info

    def render(self, mode="human", close=False):
        """
        filename_plot = "continuum_snake_velocity.png"
        plot_snake_velocity(self.data, self.period, filename_plot, 1)
        plot_curvature(self.data, self.shearable_rod.rest_lengths, self.period, 1)

        if SAVE_VIDEO:
            filename_video = "continuum_snake.mp4"
            plot_video(
                self.data,
                video_name=filename_video,
                fps=60,
                xlim=(0, 4),
                ylim=(-1, 1),
            )
        """
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering
            from gym_softrobot.utils.render.povray_renderer import Session

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rod(self.shearable_rod)

        # Temporary rendering to add side-view
        state_image = self.renderer.render(maxwidth, int(maxwidth * aspect_ratio * 0.7))
        state_image_side = self.renderer.render(
            maxwidth // 2,
            int(maxwidth * aspect_ratio * 0.3),
            camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
        )
        state_image_top = self.renderer.render(
            maxwidth // 2,
            int(maxwidth * aspect_ratio * 0.3),
            camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
        )

        state_image = np.vstack(
            [state_image, np.hstack([state_image_side, state_image_top])]
        )

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        pass

    def _build(self, include_callback: bool = False):
        # Initialize the simulation class
        snake_sim = SnakeSimulator()

        # Simulation parameters
        period = 2
        self.period = period
        final_time = (11.0 + 0.01) * period
        self.final_time = final_time
        time_step = 8e-6
        total_steps = int(final_time / time_step)
        rendering_fps = 5
        step_skip = int(1.0 / (rendering_fps * time_step))
        self.step_skip = step_skip
        callback_step_skip = int(1.0 / (60 * time_step))

        # setting up test params
        n_elem = self._n_elem
        start = np.zeros((3,))
        direction = np.array([0.0, 0.0, 1.0])
        normal = np.array([0.0, 1.0, 0.0])
        base_length = 0.35
        self.base_length = base_length
        base_radius = base_length * 0.011
        density = 1000
        nu = 1e-4
        E = 1e6
        poisson_ratio = 0.5
        shear_modulus = E / (poisson_ratio + 1.0)

        # Add rod
        self.shearable_rod = CosseratRod.straight_rod(
            n_elem,
            start,
            direction,
            normal,
            base_length,
            base_radius,
            density,
            nu,
            E,
            shear_modulus=shear_modulus,
        )
        snake_sim.append(self.shearable_rod)

        # Add gravitational forces
        gravitational_acc = -9.80665
        snake_sim.add_forcing_to(self.shearable_rod).using(
            GravityForces, acc_gravity=np.array([0.0, gravitational_acc, 0.0])
        )

        # Add muscle torques
        wave_length = 1
        muscle_torque = snake_sim.add_forcing_to(self.shearable_rod).using(
            MuscleTorques,
            base_length=base_length,
            b_coeff=np.zeros(4),
            period=period,
            wave_number=2.0 * np.pi / (wave_length),
            phase_shift=0.0,
            rest_lengths=self.shearable_rod.rest_lengths,
            ramp_up_time=period,
            direction=normal,
            with_spline=True,
        )

        # Add friction forces
        origin_plane = np.array([0.0, -base_radius, 0.0])
        normal_plane = normal
        slip_velocity_tol = 1e-8
        froude = 0.1
        mu = base_length / (period * period * np.abs(gravitational_acc) * froude)
        kinetic_mu_array = np.array(
            [mu, 1.5 * mu, 2.0 * mu]
        )  # [forward, backward, sideways]
        static_mu_array = np.zeros(kinetic_mu_array.shape)
        snake_sim.add_forcing_to(self.shearable_rod).using(
            AnisotropicFrictionalPlane,
            k=1.0,
            nu=1e-6,
            plane_origin=origin_plane,
            plane_normal=normal_plane,
            slip_velocity_tol=slip_velocity_tol,
            static_mu_array=static_mu_array,
            kinetic_mu_array=kinetic_mu_array,
        )

        # Callback
        data = defaultdict(list)
        snake_sim.collect_diagnostics(self.shearable_rod).using(
            ContinuumSnakeCallBack, step_skip=callback_step_skip, callback_params=data
        )

        # Integrator
        timestepper = PositionVerlet()
        snake_sim.finalize()
        # integrate(timestepper, snake_sim, final_time, total_steps)
        do_step, stages_and_updates = extend_stepper_interface(timestepper, snake_sim)
        stepper = partial(
            do_step,
            dt=time_step,
            TimeStepper=PositionVerlet(),
            SystemCollection=snake_sim,
            _steps_and_prefactors=stages_and_updates,
        )

        return snake_sim, stepper, muscle_torque, data



from typing import Optional

import numpy as np
import vapory

from abc import ABC, abstractmethod

from gym_softrobot.config import RendererType
from gym_softrobot.utils.render.base_renderer import BaseRenderer, BaseElasticaRendererSession

import pkg_resources

"""
For postion, change coordinate s.t.
http://www.povray.org/documentation/view/3.6.1/15/
For some reason, POVray use left-hand system
(x,y,z) -> (x,z,y)
"""

class Geom(ABC):
    @abstractmethod
    def __call__(self):
        pass

class ElasticaTexture:
    pigment = vapory.Pigment( 'color', [0.45,0.39,1.0], 'transmit', 0.0 )
    texture = vapory.Texture( pigment, vapory.Finish( 'phong', 1))

class ElasticaRod(Geom, ElasticaTexture):
    def __init__(self, rod):
        self.rod = rod

    def __call__(self):
        n_data = self.rod.n_elems+1
        pos_rad_pair = []
        pos = self.rod.position_collection 
        rad = self.rod.radius
        rad = np.concatenate([rad[0:1], 0.5*(rad[:-1]+rad[1:]), rad[-1:]])
        for i in range(n_data):
            x,z,y = pos[:,i] # transformation
            pos_rad_pair.append([x,y,z])
            pos_rad_pair.append(rad[i])

        return vapory.SphereSweep(
            'linear_spline', #'b_spline',
            n_data,
            *pos_rad_pair,
            ElasticaRod.texture
        )

class ElasticaCylinder(Geom):
    pigment = vapory.Pigment( 'color', [0.35,0.29,1.0], 'transmit', 0.0 )
    texture = vapory.Texture( pigment, vapory.Finish( 'phong', 1))

    def __init__(self, body):
        self.body = body

    def __call__(self):
        rad = self.body.radius[0]
        length = self.body.length
        tangent = self.body.director_collection[2,:,0]
        position1 = self.body.position_collection[:,0]
        position2 = position1 + length * tangent
        position1 = [position1[0], position1[2], position1[1]]
        position2 = [position2[0], position2[2], position2[1]]
        return vapory.Cylinder(
            position1,
            position2,
            rad, 
            ElasticaCylinder.texture
        )

class Sphere(Geom):
    pigment = vapory.Pigment( 'color', [1,0,1], 'transmit', 0.0 )
    texture = vapory.Texture( pigment, vapory.Finish( 'phong', 1))

    def __init__(self, loc, radius):
        x,z,y = loc
        self.sphere = vapory.Sphere(
            [x,y,z],
            radius,
            Sphere.texture
        )

    def __call__(self):
        return self.sphere

class Session(BaseElasticaRendererSession, BaseRenderer):

    def __init__(self, width, height):
        self.object_collection = []
        self.width = width
        self.height = height

        # Assets
        self.light = vapory.LightSource( [2,4,-3], 'color', [1,1,1] )

        #self.background_path = "default.inc"
        self.background_path = pkg_resources.resource_filename(
            __name__,
            'default.inc'
        )

    @property
    def type(self):
        return RendererType.POVRAY

    def add_rod(self, rod):
        self.object_collection.append(ElasticaRod(rod))

    def add_rigid_body(self, body):
        self.object_collection.append(ElasticaCylinder(body))

    def add_point(self, loc:list, radius:float):
        self.object_collection.append(Sphere(loc, radius))

    def render(self,
        width:Optional[int]=None,
        height:Optional[int]=None,
        camera_param:Optional[tuple]=None
    ):
        if not camera_param:
            #camera = vapory.Camera( 'location', [1.7,0.7,-1.2], 'look_at', [0.5,0,1] )
            #camera = vapory.Camera( 'location', [0.8,0.7,-1.2], 'look_at', [0.0,0,0] )
            camera = vapory.Camera( 'location', [0.5,0.4,-0.9], 'look_at', [0.0,0,0] )
        else:
            camera = vapory.Camera(*camera_param)
        if not width:
            width = self.width
        if not height:
            height = self.height
        objects = [obj() for obj in self.object_collection]
        objects.append(self.light)
        scene = vapory.Scene(
                    camera,
                    objects=objects,
                    included=[self.background_path])
        return scene.render(width=width, height=height)

    def close(self):
        self.object_collection.clear()



import gym 
import gym_softrobot

from gym_softrobot.config import RendererType

#gym_softrobot.RENDERER_CONFIG = RendererType.MATPLOTLIB

import argparse

def main():
    parser = argparse.ArgumentParser(description='Make registered environment and test run.')
    parser.add_argument('--env', type=str, default='OctoArmSingle-v0')
    args = parser.parse_args()

    env = gym.make(args.env, recording_fps=30)

    observation = env.reset()
    for step in range(10):
        env.render() # rendering
        input('')
        action = env.action_space.sample() 
        observation, reward, done, info = env.step(action)
        print(f"{step=:2}| {reward=}, {done=}")
        if done:
            break

if __name__=="__main__":
    main()
        




from typing import Optional

from gymnasium import spaces, Env
from gymnasium.utils import seeding

from collections import defaultdict
import time

import numpy as np


from elastica import *
from elastica.timestepper import extend_stepper_interface
from elastica._calculus import _isnan_check

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType
from gym_softrobot.utils.custom_elastica.callback_func import (
    RodCallBack,
)
from gym_softrobot.envs.octopus.controllable_constraint import ControllableFixConstraint
from gym_softrobot.envs.octopus.build import create_es_muscle_layers
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)
from gym_softrobot.utils.custom_elastica.joint import FixedJoint2Rigid
from gym_softrobot.utils.custom_elastica.constraint import BodyBoundaryCondition

from gym_softrobot.utils.actuation.actuations.muscles.muscle import ApplyMuscle


class BaseSimulator(BaseSystemCollection, Constraints, Connections, Forcing, CallBacks):
    pass


class ArmPushEnv(Env):
    """
    Description:
    Source:
    Observation:
    Actions:
    Reward:
    Starting State:
    Episode Termination:
    Solved Requirements:
    """

    metadata = {"render.modes": ["rgb_array", "human"]}

    def __init__(
        self,
        final_time: float = 2.5,
        time_step: float = 5.0e-5,
        recording_fps: int = 40,
        mode: str = "discrete",
        config_generate_video: bool = False,
        config_early_termination: bool = False,
    ):
        # Integrator type
        self.final_time = final_time
        self.time_step = time_step
        self.total_steps = int(self.final_time / self.time_step)
        self.recording_fps = recording_fps
        self.step_skip = int(1.0 / (recording_fps * time_step))

        # Simulator Config
        self.n_elem = 40

        if mode == "discrete":
            self.mode = 0
        elif mode == "continuous":
            self.mode = 1
        else:
            raise NotImplementedError(f"The mode {mode} is not available.")

        # Determinism
        seed = self.seed()

        # Action space
        if self.mode == 0:  # Discrete
            n_action = 1  # self.n_elem - 1
            # self.action_space = spaces.Box(0.0, 1.0, shape=(n_action,), dtype=np.float32)
            self.action_space = spaces.Discrete(2)

            # Observation space
            self._observation_size = (((self.n_elem + 1) * 2 + 2),)  # one hot action
            self.observation_space = spaces.Box(
                -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
            )

            self._prev_action = np.zeros(
                list(self.action_space.shape), dtype=self.action_space.dtype
            )
        elif self.mode == 1:  # Continuous
            n_action = 2
            self.action_space = spaces.Box(
                0.0, 1.0, shape=(n_action,), dtype=np.float32
            )

            # Observation space
            self._observation_size = (
                ((self.n_elem + 1) * 2 + n_action),  # one hot action
            )
            self.observation_space = spaces.Box(
                -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
            )

            self._prev_action = np.zeros(
                list(self.action_space.shape), dtype=self.action_space.dtype
            )

        # Configurations
        self.config_generate_video = config_generate_video
        self.config_early_termination = config_early_termination

        # Rendering-related
        self.viewer = None
        self.renderer = None

    def seed(self, seed=None):
        # Deprecated in new gym
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.simulator = BaseSimulator()
        self.shearable_rod = self._build()

        """ Create time stepper """
        self.StatefulStepper = PositionVerlet()
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )

        self.time = np.float64(0.0)

        # Initial State
        state = self.get_state()

        if return_info:
            return state, {}
        else:
            return state

    def _build(self):
        """Set up arm params"""
        n_elem = self.n_elem
        L0 = 0.2
        radius_base = 0.012  # radius of the arm at the base
        radius_tip = 0.001  # radius of the arm at the tip
        radius = np.linspace(radius_base, radius_tip, n_elem + 1)
        radius_mean = (radius[:-1] + radius[1:]) / 2
        damp_coefficient = 0.05 * 2

        shearable_rod = CosseratRod.straight_rod(
            n_elements=n_elem,
            start=np.zeros((3,)),
            direction=np.array([1.0, 0.0, 0.0]),
            normal=np.array([0.0, 1.0, -0.0]),
            base_length=L0,
            base_radius=radius_mean.copy(),
            density=700,
            nu=damp_coefficient * ((radius_mean / radius_base) ** 2) * 1e2,
            youngs_modulus=1e4,
            shear_modulus=1e4 / 1.5,
            # poisson_ratio=0.5,
            # nu_for_torques=damp_coefficient * ((radius_mean / radius_base) ** 4),
        )
        self.simulator.append(shearable_rod)

        controller_id = (
            self.simulator.constrain(shearable_rod)
            .using(
                ControllableFixConstraint,
                index=0,
            )
            .id()
        )

        """ Add muscle actuation """
        self.muscle_layers = create_es_muscle_layers(radius_mean, radius_base)

        self.muscles_parameters = []
        if self.config_generate_video:
            for _ in self.muscle_layers:
                self.muscles_parameters.append(defaultdict(list))

        self.simulator.add_forcing_to(shearable_rod).using(
            ApplyMuscle,
            muscles=self.muscle_layers,
            step_skip=self.step_skip,
            callback_params_list=self.muscles_parameters,
        )

        # CallBack
        if self.config_generate_video:
            self.rod_parameters_dict = defaultdict(list)
            self.simulator.collect_diagnostics(shearable_rod).using(
                RodCallBack,
                step_skip=self.step_skip,
                callback_params=self.rod_parameters_dict,
            )

        self.simulator.finalize()

        controllable_constraint = dict(self.simulator._constraints)[controller_id]
        self.BC = controllable_constraint.get_controller
        self.BC.turn_on()

        return shearable_rod

    def get_state(self):
        # Build state
        rod = self.shearable_rod
        pos_state1 = rod.position_collection[0]  # x
        vel_state1 = rod.velocity_collection[0]  # x
        # pos_state2 = rod.position_collection[1]  # y
        previous_action = self._prev_action
        if self.mode == 0:  # Discrete
            state = np.hstack(
                [pos_state1, vel_state1, np.eye(2)[previous_action]]
            ).astype(np.float32)
        elif self.mode == 1:  # Continuous
            state = np.hstack(
                [
                    pos_state1,
                    vel_state1,
                    previous_action,
                ]
            ).astype(np.float32)
        return state

    def set_action(self, action) -> None:
        scale = 1.0  # min(time / 0.02, 1)

        # Continuous action
        # for muscle_count, muscle_layer in enumerate(self.muscle_layers):
        #    muscle_layer.set_activation(action[muscle_count] * scale)

        if self.mode == 0:  # Discrete Action
            if action == 0:  # Fix last node and activate muscle
                self.BC.index = 0
                self.muscle_layers[0].set_activation(-0.0 * scale)
                self.muscle_layers[1].set_activation(0.0 * scale)
                self.muscle_layers[2].set_activation(0.5 * scale)
            elif action == 1:  # Fix first node and release muscle
                self.BC.index = -1
                self.muscle_layers[0].set_activation(0.0)
                self.muscle_layers[1].set_activation(0.0)
                self.muscle_layers[2].set_activation(0.0)
            else:
                raise NotImplementedError("Action must be 1 or 0")
        elif self.mode == 1:  # Continuous Action
            location = action[0]
            activation = action[1]
            self.BC.index = int(np.clip(location * self.n_elem, 0, self.n_elem - 1))
            self.muscle_layers[2].set_activation(activation)

        # update previous action
        self._prev_action = action

    def step(self, action):
        """Set intrinsic strains (set actions)"""
        self.set_action(action)

        """ Post-simulation """
        prev_cm_pos = self.shearable_rod.compute_position_center_of_mass()[:2]

        """ Run the simulation for one step """
        stime = time.perf_counter()
        for _ in range(self.step_skip):
            self.time = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time,
                self.time_step,
            )
        etime = time.perf_counter()

        """ Done is a boolean to reset the environment before episode is completed """
        cm_pos = self.shearable_rod.compute_position_center_of_mass()[:2]

        terminated = False
        truncated = False
        survive_reward = 0.0
        forward_reward = 0.0
        # Position of the rod cannot be NaN, it is not valid, stop the simulation
        invalid_values_condition = _isnan_check(
            np.concatenate(
                [
                    self.shearable_rod.position_collection.ravel(),
                    self.shearable_rod.velocity_collection.ravel(),
                    self.shearable_rod.director_collection.ravel(),
                    self.shearable_rod.alpha_collection.ravel(),
                    self.shearable_rod.omega_collection.ravel(),
                    cm_pos.ravel(),
                ]
            )
        )

        if self.config_early_termination:
            terminated = self.check_early_termination()
            truncated = terminated
            survive_reward = -10.0
        elif invalid_values_condition:
            # print(f" Nan detected in, exiting simulation now. {self.time=}")
            terminated = True
            truncated = True
            survive_reward = -20.0
        else:
            moved_distance = np.linalg.norm(cm_pos, ord=2) - np.linalg.norm(
                prev_cm_pos, ord=2
            )
            forward_reward = moved_distance

        """ Time limit """
        timelimit = False
        if self.time > self.final_time:
            # survive_reward = np.linalg.norm(cm_pos, ord=2) * 10
            timelimit = True
            terminated = True

        reward = forward_reward + survive_reward
        if np.isnan(reward):
            terminated = True
            truncated = True
            reward = -20.0
        # reward *= 10 # Reward scaling
        # print(f'{reward=:.3f}: {forward_reward=:.3f}, {survive_reward=:.3f}')

        """ Return state:
            (1) current simulation time
            (2) current systems
            (3) a flag denotes whether the simulation runs correlectly
        """
        # systems = [self.shearable_rod]
        states = self.get_state()
        if np.any(np.isnan(states)):
            terminated = True
            reward = -20.0
            states = np.nan_to_num(states)

        # Info
        info = {
            "time": self.time,
            "rod": self.shearable_rod,
            "TimeLimit.truncated": timelimit,
        }

        return states, reward, terminated, truncated, info

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rod(self.shearable_rod)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            # TODO Maybe add 2D rendering instead
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None

    # def save_data(self, dir, eps):

    #     print("Saving data to pickle files...")

    #     import pickle

    #     with open(dir + "/simulation_data%03d.pickle" % eps, "wb") as f:
    #         data = dict(
    #             rods=[self.rod_parameters_dict], muscles=self.muscles_parameters
    #         )
    #         pickle.dump(data, f)

    #     with open(dir + "/simulation_systems%03d.pickle" % eps, "wb") as f:
    #         data = dict(rods=[self.shearable_rod], muscles=self.muscle_layers)
    #         pickle.dump(data, f)

    def check_early_termination(self, cutoff_error=1e-7):
        desired_Hamiltonian = self.cal_desired_Hamiltonian()
        if desired_Hamiltonian < cutoff_error:
            return True
        return False

    def cal_desired_Hamiltonian(self):
        kinetic_energy = (
            self.shearable_rod.compute_translational_energy()
            + self.shearable_rod.compute_rotational_energy()
        )
        desired_potential_energy = (
            self.shearable_rod.compute_shear_energy()
            + self.shearable_rod.compute_bending_energy()
        )
        return kinetic_energy + desired_potential_energy

    # def post_processing(self, algo_data, cutoff_error=1e-7):
    #     for k in range(len(self.rod_parameters_dict["time"])):

    #         # calculate the desired Hamiltonian for every time frame
    #         kinetic_energy = compute_translational_energy(
    #             self.shearable_rod.mass, self.rod_parameters_dict["velocity"][k]
    #         ) + compute_rotational_energy(
    #             self.shearable_rod.mass_second_moment_of_inertia,
    #             self.rod_parameters_dict["omega"][k],
    #             self.rod_parameters_dict["dilatation"][k],
    #         )
    #         desired_potential_energy = compute_shear_energy(
    #             self.rod_parameters_dict["sigma"][k],
    #             algo_data["sigma"],
    #             self.shearable_rod.shear_matrix,
    #             self.rod_parameters_dict["dilatation"][k]
    #             * self.shearable_rod.rest_lengths,
    #         ) + compute_bending_energy(
    #             self.rod_parameters_dict["kappa"][k],
    #             algo_data["kappa"],
    #             self.shearable_rod.bend_matrix,
    #             self.rod_parameters_dict["voronoi_dilatation"][k]
    #             * self.shearable_rod.rest_voronoi_lengths,
    #         )
    #         desired_Hamiltonian = kinetic_energy + desired_potential_energy

    #         # calculate the control energy
    #         self.muscles_parameters[0]["control_energy"].append(
    #             0.5 * np.sum(self.muscles_parameters[0]["activation"][k] ** 2)
    #         )

    #         # check if the desired Hamiltonian is smaller then cutoff error
    #         flag = False
    #         if desired_Hamiltonian < cutoff_error:
    #             flag = True
    #         self.rod_parameters_dict["stable_flag"].append(flag)

    #     for k in range(len(self.rod_parameters_dict["time"])):
    #         if self.rod_parameters_dict["stable_flag"][-k - 1] is False:
    #             self.rod_parameters_dict["stable_flag"][: -k - 1] = [
    #                 False
    #                 for kk in range(
    #                     len(self.rod_parameters_dict["stable_flag"][: -k - 1])
    #                 )
    #             ]
    #             break

    #     energy_cost = 0
    #     reach_time = self.rod_parameters_dict["time"][-1]
    #     for k in range(len(self.rod_parameters_dict["time"])):
    #         if self.rod_parameters_dict["stable_flag"][k] is False:
    #             energy_cost += self.muscles_parameters[0]["control_energy"][k]
    #         else:
    #             reach_time = self.rod_parameters_dict["time"][k]

    #     return reach_time, energy_cost


class ArmPullWeightEnv(ArmPushEnv):
    def __init__(self, **kwargs):
        super().__init__(time_step=2.5e-5, **kwargs)

    def _build(self):
        """Set up arm params"""
        n_elem = self.n_elem
        L0 = 0.2
        radius_base = 0.012  # radius of the arm at the base
        radius_tip = 0.001  # radius of the arm at the tip
        radius = np.linspace(radius_base, radius_tip, n_elem + 1)
        radius_mean = (radius[:-1] + radius[1:]) / 2
        density = 700
        damp_coefficient = 0.05 * 2

        shearable_rod = CosseratRod.straight_rod(
            n_elements=n_elem,
            start=np.zeros((3,)),
            direction=np.array([1.0, 0.0, 0.0]),
            normal=np.array([0.0, 1.0, -0.0]),
            base_length=L0,
            base_radius=radius_mean.copy(),
            density=density,
            nu=damp_coefficient * ((radius_mean / radius_base) ** 2) * 5e2,
            youngs_modulus=1e4,
            shear_modulus=1e4 / 1.5,
            # poisson_ratio=0.5,
            # nu_for_torques=damp_coefficient * ((radius_mean / radius_base) ** 4),
        )
        self.simulator.append(shearable_rod)

        """ Add head """
        rigid_rod_length = radius_base * 2
        rigid_rod_radius = 0.015
        start = np.zeros((3,))
        start[0] = -rigid_rod_radius * 0.9
        start[2] = -2 * radius_base
        direction = np.array([0.0, 0.0, 1.0])
        normal = np.array([0.0, 1.0, 0.0])

        rigid_rod = Cylinder(
            start, direction, normal, rigid_rod_length, rigid_rod_radius, density * 1.0
        )
        self.simulator.append(rigid_rod)
        self.rigid_rod = rigid_rod

        """ Constraint body """
        self.simulator.constrain(rigid_rod).using(
            # Upright rigid rod need restoration force/torque against the floor
            BodyBoundaryCondition,
            constrained_position_idx=(0,),
            constrained_director_idx=(0,),
        )

        """ Set up boundary conditions """
        _k = 1e6
        _kt = 1e0
        self.simulator.connect(
            first_rod=rigid_rod,
            second_rod=shearable_rod,
            first_connect_idx=-1,
            second_connect_idx=0,
        ).using(
            FixedJoint2Rigid, k=_k, nu=1e-2, kt=_kt, angle=0, radius=rigid_rod_radius
        )

        """ Set up controller """
        controller_id = (
            self.simulator.constrain(shearable_rod)
            .using(ControllableFixConstraint, index=0, reduction_ratio=0.9)
            .id()
        )

        """ Add muscle actuation """
        self.muscle_layers = create_es_muscle_layers(radius_mean, radius_base)

        self.muscles_parameters = []

        self.simulator.add_forcing_to(shearable_rod).using(
            ApplyMuscle,
            muscles=self.muscle_layers,
            step_skip=self.step_skip,
            callback_params_list=self.muscles_parameters,
        )

        # CallBack
        if self.config_generate_video:
            self.rod_parameters_dict = defaultdict(list)
            self.simulator.collect_diagnostics(shearable_rod).using(
                RodCallBack,
                step_skip=self.step_skip,
                callback_params=self.rod_parameters_dict,
            )

        self.simulator.finalize()

        controllable_constraint = dict(self.simulator._constraints)[controller_id]
        self.BC = controllable_constraint.get_controller
        self.BC.turn_on()

        return shearable_rod

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rod(self.shearable_rod)
            self.renderer.add_rigid_body(self.rigid_rod)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            # TODO Maybe add 2D rendering instead
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image



"""
Created on Mar. 11, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import matplotlib.colors as mcolors

default_colors = [
    '#1f77b4',  # muted blue
    '#ff7f0e',  # safety orange
    '#2ca02c',  # cooked asparagus green
    '#d62728',  # brick red
    '#9467bd',  # muted purple
    '#8c564b',  # chestnut brown
    '#e377c2',  # raspberry yogurt pink
    '#7f7f7f',  # middle gray
    '#bcbd22',  # curry yellow-green
    '#17becf'   # blue-teal
    ]
    
default_label_fontsize = 12
paper_label_fontsize = 48
paper_linewidth = 5

rod_color = mcolors.BASE_COLORS['m']
algo_color = mcolors.BASE_COLORS['g']

def change_box_to_arrow_axes(fig, ax, linewidth=1.0, overhang=0.0, xaxis_ypos=0, yaxis_xpos=0, x_offset=[0, 0], y_offset=[0, 0]):
    for spine in ax.spines.values():
        spine.set_visible(False)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    # get width and height of axes object to compute
    # matching arrowhead length and width
    dps = fig.dpi_scale_trans.inverted()
    bbox = ax.get_window_extent().transformed(dps)
    width, height = bbox.width, bbox.height

    # manual arrowhead width and length
    hw = 1./40.*(ymax-ymin)
    hl = 1./40.*(xmax-xmin)
    lw = linewidth # axis line width
    ohg = overhang # arrow overhang

    # compute matching arrowhead length and width
    yhw = hw/(ymax-ymin)*(xmax-xmin)* height/width
    yhl = hl/(xmax-xmin)*(ymax-ymin)* width/height

    # draw x and y axis
    ax.arrow(xmin+x_offset[0], xaxis_ypos, xmax-xmin+x_offset[1], 0, fc='k', ec='k', lw = lw, 
                head_width=hw, head_length=hl, overhang = ohg, 
                length_includes_head= True, clip_on = False) 

    ax.arrow(yaxis_xpos, ymin+y_offset[0], 0, ymax-ymin+y_offset[1], fc='k', ec='k', lw = lw, 
                head_width=yhw, head_length=yhl, overhang = ohg, 
                length_includes_head= True, clip_on = False)
    return ax


import pytest
import numpy as np

from gym import envs
from gym.spaces import Box
from gym.utils.env_checker import check_env

from tests.envs.spec_list import spec_list

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType

#RENDERER_CONFIG = RendererType.POVRAY
RENDERER_CONFIG = RendererType.MATPLOTLIB

# FIXME: Testing rendering is disabled for now. It needs to install POVRAY on CI
#@pytest.mark.parametrize("spec", spec_list)
#def test_env_render_result_np_array_for_rgb_mode(spec):
#    env = spec.make()
#    env.reset()
#    output = env.render(mode='rgb_array')
#    assert isinstance(output, np.ndarray)
#    assert output.shape[2] == 3
#    env.close()



"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit
from tqdm import tqdm

from collections import defaultdict
import pickle

import matplotlib.pyplot as plt
from matplotlib import gridspec
from mpl_toolkits.mplot3d import Axes3D

from elastica._linalg import _batch_cross, _batch_matvec
from elastica._calculus import quadrature_kernel
from elastica.rod.cosserat_rod import CosseratRod

from gym_softrobot.utils.actuation.algorithms.algorithm import Algorithm
from gym_softrobot.utils.actuation.objects.point import PointTarget
from gym_softrobot.utils.actuation.rod_tools import forward_path, calculate_dilatation#, calculate_potential_energy

label_fontsize = 14

def strain_to_shear_and_curvature(strain):
    n_elems = int((strain.size + 3) / 6)
    shear = np.zeros((3, n_elems))
    curvature = np.zeros((3, n_elems-1))
    index = 0
    for i in range(3):
        shear[i, :] = strain[index:index+n_elems]
        index += n_elems
    for i in range(3):
        curvature[i] = strain[index:index+n_elems-1]
        index += (n_elems-1)
    return shear, curvature

def shear_and_curvature_to_strain(shear, curvature):
    strain = np.zeros(shear.size + curvature.size)
    n_elems = shear.shape[1]
    index = 0
    for i in range(3):
        strain[index:index+n_elems] = shear[i, :]
        index += n_elems
    for i in range(3):
        strain[index:index+n_elems-1] = curvature[i]
        index += (n_elems-1)
    return strain

@njit(cache=True)
def inverse_rigidity_matrix(matrix):
    inverse_matrix = np.zeros(matrix.shape)
    for i in range(inverse_matrix.shape[2]):
        inverse_matrix[:, :, i] = np.linalg.inv(matrix[:, :, i])
    return inverse_matrix

@njit(cache=True)
def backward_path(dl, director, shear, curvature, internal_force, internal_couple):
    blocksize = internal_force.shape[1]
    for i in range(blocksize-1):
        internal_force[:, -1-i-1] = internal_force[:, -1-i] + (
            np.cross(curvature[:, -1-i], internal_force[:, -1-i])
            ) * dl
    nu_cross_n = _batch_cross(shear, internal_force)

    for i in range(blocksize-2):
        kappa_cross_m = np.cross(curvature[:, -1-i], internal_couple[:, -1-i])
        internal_couple[:, -1-i-1] = internal_couple[:, -1-i] + (
            kappa_cross_m
            + (nu_cross_n[:, -1-i] + nu_cross_n[:, -1-i-1])/2
            ) * dl
    pass

@njit(cache=True)
def update_strain(shear, curvature, internal_force, internal_couple, shear_matrix, bend_matrix, eta_shear, eta_curvature):
    shear_intrinsic = np.zeros(shear.shape)
    shear_intrinsic[2, :] = 1
    shear[:, :] += eta_shear * (internal_force - _batch_matvec(shear_matrix, shear-shear_intrinsic))
    shear[2, :] = np.maximum(shear[2, :], 0)
    curvature[:, :] += eta_curvature * (internal_couple - _batch_matvec(bend_matrix, curvature))
    pass

class ForwardBackward(Algorithm):
    def __init__(self, rod, algo_config):
        Algorithm.__init__(self, rod, algo_config)
        self.eta_shear = np.zeros(self.shear.shape)
        self.eta_curvature = np.zeros(self.curvature.shape)

        min_rigidity = np.inf
        for i in range(3):
            min_rigidity_i = min(np.min(self.shear_matrix[i, i, :]), np.min(self.shear_matrix[i, i, :]))
            min_rigidity = min(min_rigidity, min_rigidity_i)
            
        for i in range(3):
            self.eta_shear[i, :] = algo_config.get('eta', 1)/(self.shear_matrix[i, i, :]/min_rigidity)
            self.eta_curvature[i, :] = algo_config.get('eta', 1)/(self.bend_matrix[i, i, :]/min_rigidity)

    def update(self,):
        forward_path(self.dl, self.shear, self.curvature, self.position, self.director)
        terminal_cost_partial_end_position, terminal_cost_partial_end_director = (
            self.target.terminal_cost_partial_end_state(
                self.position[:, -1],
                end_director=self.director[:, :, -1]
            )
        )
        # print(terminal_cost_partial_end_director)
        # print(self.position)

        self.internal_force[:, -1] = - self.director[:, :, -1] @ terminal_cost_partial_end_position
        self.internal_couple[:, -1] = terminal_cost_partial_end_director
        backward_path(self.dl, self.director, self.shear, self.curvature, self.internal_force, self.internal_couple)
        
        update_strain(
            self.shear, self.curvature,
            self.internal_force, self.internal_couple,
            self.shear_matrix, self.bend_matrix,
            self.eta_shear, self.eta_curvature
        )
        
        pass

    def run(self, continuum_target_cost_weight=None, plot_flag=False, iter_number=100000):
        for _ in tqdm(range(iter_number)):
            self.update()
        return

def reach_test_function(
        sampling_number=1000,
        iter_number=600000,
        error_tolerance=1e-6
    ):
    n = 100
    L0 = 0.2

    """ Set up arm params """
    radius_base = 0.012  # radius of the arm at the base
    radius_tip = 0.012  # radius of the arm at the tip
    # radius_tip = 0.001  # radius of the arm at the tip
    radius = np.linspace(radius_base, radius_tip, n+1)
    radius_mean = (radius[:-1]+radius[1:])/2
    damp_coefficient = 0.03

    rod = CosseratRod.straight_rod(
        n_elements=n,
        start=np.zeros((3,)),
        direction=np.array([1.0, 0.0, 0.0]),
        normal=np.array([0.0, 0.0, -1.0]),
        base_length=L0,
        base_radius=radius_mean.copy(),
        density=700,
        nu=damp_coefficient*((radius_mean/radius_base)**2),
        youngs_modulus=1e4,
        poisson_ratio=0.5,
        nu_for_torques=damp_coefficient*((radius_mean/radius_base)**4),
    )

    target_position = np.array([0.5, 0.5, 0])
    target_director = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])
    algo_config = dict(
        target=PointTarget(
            position=target_position*L0,
            director=target_director,
            target_cost_weight=100000),
        eta=1e-10,
    )

    algo = ForwardBackward(rod, algo_config)

    # target_cost_weight_array = np.linspace(1000000, 1000000, 1)
    target_cost_weight_array = np.logspace(0, 6, 6)

    fig = plt.figure(figsize=(12, 8))
    gs = gridspec.GridSpec(
        nrows=3, ncols=4, figure=fig, width_ratios=[1, 1, 1, 1], height_ratios=[1, 1, 1]
    )
    # ax_2d = fig.add_subplot(gs[0:2, 0:2])
    ax_3d = fig.add_subplot(gs[0:2, 0:2], projection='3d')
    ax_J1 = fig.add_subplot(gs[2, 0])
    ax_J2 = fig.add_subplot(gs[2, 1])
    ax_shear = []
    ax_curvature = []
    for index_i in range(3):
        ax_shear.append(fig.add_subplot(gs[index_i, 2], xlim=[-0.1, 1.1]))
        ax_curvature.append(fig.add_subplot(gs[index_i, 3], xlim=[-0.1, 1.1]))

    # alphas = np.zeros(algo.alpha.shape + target_cost_weight_array.shape)
    running_costs = np.zeros(target_cost_weight_array.shape)
    terminal_costs = np.zeros(target_cost_weight_array.shape)
    # print(algo.position)
    # algo.run(ax_2d, iter_number=100000)
    # print(algo.position)
    for i in tqdm(range(iter_number)):
        # print("123456")
        if (i % sampling_number) == 0:
            ax_3d.plot(
                algo.position[0, :]/L0,
                algo.position[1, :]/L0,
                algo.position[2, :]/L0,
                color='#1f77b4',
                alpha=i/iter_number
            )
            # ax_2d.plot(
            #     algo.position[0, :]/L0,
            #     algo.position[1, :]/L0,
            #     color='#1f77b4',
            #     alpha=i/iter_number
            # )
            # force = -algo.target.terminal_cost_partial_end_state(algo.position[:, -1], algo.director[:, :, -1])
            # v = np.zeros((3, 2))
            # v[:, 0] = algo.position[:, -1]/L0
            # v[:, 1] = algo.position[:, -1]/L0 + force/np.linalg.norm(force)*0.1
            # ax_2d.plot(v[0, :], v[1, :])
            for index_i in range(3):
                ax_shear[index_i].plot(algo.s_shear, algo.shear[index_i, :], color='#1f77b4', alpha=i/iter_number)
                ax_curvature[index_i].plot(algo.s_curvature, algo.curvature[index_i, :], color='#1f77b4', alpha=i/iter_number)
                # print(np.max(algo.curvature[2, :]), np.min(algo.curvature[2, :]))
        # if i==2 :
        #     quit()
        algo.update()


    # for cost_weight_i, target_cost_weight in enumerate(target_cost_weight_array):
    #     algo_config['target'].target_cost_weight = target_cost_weight
    #     algo.run()
    #     # alphas[:, cost_weight_i] = algo.alpha.copy()

    #     running_costs[cost_weight_i] = algo.running_cost
    #     terminal_costs[cost_weight_i] = algo.terminal_cost
        
    #     ax_2d.plot(algo.position_collection[0, :]/L0, algo.position_collection[1, :]/L0)
    #     for index_i in range(3):
    #         ax_strain[index_i].plot(algo.s_strain, algo.strain[index_i, :])
    #         ax_curvature[index_i].plot(algo.s_curvature, algo.curvature[index_i, :])

    # for alpha_i in range(algo.alpha.shape[0]):
    #     ax_J1.plot(target_cost_weight_array, alphas[alpha_i], label=r'$\alpha_{}$'.format(alpha_i+1))
    ax_J1.legend()
    ax_J2.semilogy(target_cost_weight_array, running_costs+terminal_costs, label=r'$\mathsf{J}$', color='black')
    ax_J2.semilogy(target_cost_weight_array, running_costs, label='$V$', color='black', linestyle='--')
    ax_J2.semilogy(target_cost_weight_array, terminal_costs, label='$\Phi$', color='black', linestyle='dotted')
    ax_J2.legend()
    
    y_lim = [-1.1, 1.1]
    ax_J1.set_ylim(y_lim)
    
    # y_lim = np.array(ax_J2.get_ylim())
    # y_lim[0] = min(np.min(running_costs[1:]),np.min(terminal_costs[1:]))
    y_lim = np.array([0.009, 2.1])
    ax_J2.set_ylim(y_lim)
    fig.text(0.3, 0.04, "continuation parameter $\mu$", ha='center', va='center', fontsize=label_fontsize)
    fig.text(0.71, 0.04, '$s$', ha='center', va='center', fontsize=label_fontsize)

    ax_3d.scatter(
        algo.target.position[0]/L0,
        algo.target.position[1]/L0,
        algo.target.position[2]/L0,
        c='r', marker='x'
    )
    # ax_2d.scatter(
    #     algo.target.position[0]/L0,
    #     algo.target.position[1]/L0,
    #     c='r', marker='x'
    # )
    # ax_2d.set_xlim([-0.1, 1.1])
    # ax_2d.set_ylim([-0.1, 1.1])
    ax_3d.set_xlim([-0.1, 1.1])
    ax_3d.set_ylim([-0.1, 1.1])
    ax_3d.set_zlim([-0.6, 0.6])
    ax_shear[0].set_title('shear', fontsize=label_fontsize)
    ax_curvature[0].set_title('curvature', fontsize=label_fontsize)
    for index_i in range(3):
        ax_curvature[index_i].text(1.2, 0, '$d_{}$'.format(index_i+1), ha='center', va='center', fontsize=label_fontsize)
    for index_i in range(2):
        ax_shear[index_i].set_ylim([-0.11, 0.11])
        ax_shear[index_i].ticklabel_format(axis='y', scilimits=(-1, -1))  
        ax_curvature[index_i].set_ylim([-110, 110])
        ax_curvature[index_i].ticklabel_format(axis='y', scilimits=(2, 2))
    ax_shear[2].set_ylim([-0.1, 2.1])
    ax_shear[2].text(0, 2, 'stretch', fontsize=label_fontsize, ha='left', va='top')
    ax_curvature[2].set_ylim([-11, 11])
    ax_curvature[2].ticklabel_format(axis='y', scilimits=(1, 1))
    ax_curvature[2].text(1, 10, 'twist', fontsize=label_fontsize, ha='right', va='top')
    plt.show()

if __name__ == "__main__":
    reach_test_function()
